/****************************************************************************
* VCGLib                                                            o o     *
* Visual and Computer Graphics Library                            o     o   *
*                                                                _   O  _   *
* Copyright(C) 2004-2012                                           \/)\/    *
* Visual Computing Lab                                            /\/|      *
* ISTI - Italian National Research Council                           |      *
*                                                                    \      *
* All rights reserved.                                                      *
*                                                                           *
* This program is free software; you can redistribute it and/or modify      *
* it under the terms of the GNU General Public License as published by      *
* the Free Software Foundation; either version 2 of the License, or         *
* (at your option) any later version.                                       *
*                                                                           *
* This program is distributed in the hope that it will be useful,           *
* but WITHOUT ANY WARRANTY; without even the implied warranty of            *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the             *
* GNU General Public License (http://www.gnu.org/licenses/gpl.txt)          *
* for more details.                                                         *
*                                                                           *
****************************************************************************/

#include <vcg/complex/complex.h>
#include <wrap/io_trimesh/import_ply.h>
#include <wrap/io_trimesh/export_ply.h>
#include <vcg/complex/algorithms/update/color.h>
#include <vcg/complex/algorithms/geodesic.h>
#include <algorithm>
using namespace vcg;
using namespace std;


class MyEdge;
class MyFace;
class MyVertex;
struct MyUsedTypes : public vcg::UsedTypes<	vcg::Use<MyVertex>   ::AsVertexType,
                                        vcg::Use<MyEdge>     ::AsEdgeType,
                                        vcg::Use<MyFace>     ::AsFaceType>{};

class MyVertex  : public vcg::Vertex<MyUsedTypes, vcg::vertex::Coord3f, vcg::vertex::VFAdj, vertex::Qualityf, vcg::vertex::Color4b, vcg::vertex::Normal3f, vcg::vertex::BitFlags  >{};
class MyFace    : public vcg::Face<  MyUsedTypes, vcg::face::VFAdj,  vcg::face::VertexRef, vcg::face::FFAdj, vcg::face::FEAdj, vcg::face::BitFlags > {};
class MyEdge    : public vcg::Edge<  MyUsedTypes, vcg::edge::VertexRef, vcg::edge::BitFlags>{};
class MyMesh    : public vcg::tri::TriMesh< std::vector<MyVertex>, std::vector<MyFace> , std::vector<MyEdge>  > {};

class GeodesicPos
{
public:
  face::Pos<MyFace> p;
  float dist;

  bool operator < (const GeodesicPos & p) const
  {
    return dist > p.dist;
  }

  GeodesicPos(MyFace *f, int edgeInd)
  {
    p = face::Pos<MyFace>(f,edgeInd);
    face::Pos<MyFace> tp= p;
    tp.FlipF();
    tp.FlipE();
    tp.FlipV();
    dist = tp.V()->Q();
  }
};

void Dump(MyMesh &m, const char *msg, int ind)
{
  printf("Saving %s...\n",msg);
  char buf[128]; sprintf(buf,msg,ind);
  MyMesh tmpEdge;
  for(MyMesh::EdgeIterator ei=m.edge.begin();ei!=m.edge.end();++ei)
  {
    if(ei->IsS())
    {
      MyMesh::VertexIterator new_vi = tri::Allocator<MyMesh>::AddVertices(tmpEdge,2);
      MyMesh::EdgeIterator new_ei = tri::Allocator<MyMesh>::AddEdges(tmpEdge,1);
      new_ei->V(0)=&*new_vi;
      new_vi->P() = ei->V(0)->P();
      new_vi++;
      new_ei->V(1)=&*new_vi;
      new_vi->P() = ei->V(1)->P();
    }
  }
  tri::io::ExporterPLY<MyMesh>::Save(tmpEdge,buf,tri::io::Mask::IOM_VERTCOLOR|tri::io::Mask::IOM_VERTQUALITY|tri::io::Mask::IOM_EDGEINDEX,false);
}

/* Counts and returns the number of selected edges incident on a vertex vp;
 * it returns in ep the pointer to one of the selected edge (if any)
 */
int CountSelectedEdges(MyVertex *vp, MyEdge *&ep)
{
  ep=0;
  MyFace *fp = vp->VFp();
  face::Pos<MyFace> startPos(fp,vp);
  face::Pos<MyFace> curPos=startPos;
  int selNum=0;
  do
  {
    curPos.FlipE();
    curPos.FlipF();
    if(curPos.F()->FEp(curPos.E())->IsS())
    {
      ep = curPos.F()->FEp(curPos.E());
      selNum++;
    }
  }
  while(curPos!=startPos);
  return selNum;
}
/* Counts and returns the number of selected edges incident on a face fp;
 * If there are exactly two selected edges and both of them are not adjacent to a vertex with more than two incident selected edges
 * it returns in ep the pointer to only the un-selected edge.
 * Used to 'smooth' the chain of selected edges.
 */

int CountSelectedEdges(MyFace *fp, int &eInd)
{
  int unselNum=0;
  for(int i=0;i<3;++i)
  {
    MyEdge *ep;
    if(CountSelectedEdges(fp->V(i),ep)>2)
      return -1;

    if(!fp->FEp(i)->IsS())
    {
      unselNum++;
      eInd = i;
    }
  }
  return 3-unselNum;
}
// Aux Per-face class that we use when we cut the mesh.
// for each wedge of the face we keep the index of the vertex to be added
class NewVertInd
{
public:
  int v[3];
  NewVertInd(){v[0]=-1;v[1]=-1;v[2]=-1;}
};

int main()
{
  MyMesh m;
  tri::io::ImporterPLY<MyMesh>::Open(m,"fertility10k.ply");
//  tri::io::ImporterPLY<MyMesh>::Open(m,"torus_irregular.ply");
//  tri::io::ImporterPLY<MyMesh>::Open(m,"2handle.ply");

  // Step 1)
  // Choose a face and store geodesic distance from the vertexes of this face to all the other vertexes of the mesh.
  vector<MyMesh::VertexPointer> seedVec;
  MyMesh::FacePointer startFace=&m.face[0];
  seedVec.push_back(startFace->V(0));
  seedVec.push_back(startFace->V(1));
  seedVec.push_back(startFace->V(2));
  tri::UpdateTopology<MyMesh>::VertexFace(m);
  tri::Geodesic<MyMesh>::Compute(m,seedVec);
  tri::UpdateColor<MyMesh>::PerVertexQualityRamp(m);
  tri::io::ExporterPLY<MyMesh>::Save(m,"out.ply",tri::io::Mask::IOM_VERTCOLOR|tri::io::Mask::IOM_VERTQUALITY);
  tri::UpdateTopology<MyMesh>::FaceFace(m);
  tri::UpdateTopology<MyMesh>::AllocateEdge(m);

  // Step 1)
  // Make a topologic visit of the mesh in distance order.
  // Use a priority queue ordered according to the previously computed geodesic distance.
  // Once two faces expand onto a common edge select this edge;
  // In this way you build a cutting tree that unfold the mesh into a disk

  vector<GeodesicPos> heap;
  heap.push_back(GeodesicPos(startFace,0));
  heap.push_back(GeodesicPos(startFace,1));
  heap.push_back(GeodesicPos(startFace,2));

  make_heap(heap.begin(),heap.end());

  tri::UpdateSelection<MyMesh>::FaceClear(m);
  startFace->SetS();
  int cnt=0;
  while(!heap.empty())
  {
    pop_heap(heap.begin(),heap.end());
    GeodesicPos cur = heap.back();
    heap.pop_back();
    MyFace *curFace = cur.p.FFlip();
    if(!curFace->IsS())
    {
      curFace->SetS();
      for(int i=0;i<3;++i)
      {
        if(!curFace->FFp(i)->IsS())
          heap.push_back(GeodesicPos(curFace,i));
      }
      make_heap(heap.begin(),heap.end());
    }
    else
    {
      MyEdge *ep = cur.p.F()->FEp(cur.p.E());
      ep->SetS();
    }
    if(cnt++%5000 == 0)
      Dump(m,"edge%03i.ply",cnt);
  }

  // Step 3)
  // Prune the tree, removing all dangling edges and leaving only the non-tree component (loops)
  // We use a stack of active vertices and deselect all dangling edges

  cnt=0;
  vector<MyVertex *> pruneStack;
  for(MyMesh::VertexIterator vi=m.vert.begin();vi!=m.vert.end();vi++)
    pruneStack.push_back(&*vi);

  while(!pruneStack.empty())
  {
    MyVertex *vp = pruneStack.back();
    pruneStack.pop_back();
    MyEdge *ep=0;
    if(CountSelectedEdges(vp,ep)==1)
    {
      assert(ep!=0);
      ep->ClearS();
      if(ep->V(0)==vp) pruneStack.push_back(ep->V(1));
      else pruneStack.push_back(ep->V(0));
    }
    if(cnt++%5000 == 0) Dump(m,"pruning%05i.ply",cnt);
  }
  Dump(m,"final.ply",0);

  // Step 4)
  // Trivial path smoothing
  // Shortcut each pair of edges as below
  //        /\
  //       /  \      --INTO-->
  //      /    \
  // ___./      \._____    _____.________.______
  //
  cnt=0;
  vector<MyFace *> zigzagStack;
  for(MyMesh::FaceIterator fi=m.face.begin();fi!=m.face.end();fi++)
    zigzagStack.push_back(&*fi);
  while(!zigzagStack.empty())
  {
    MyFace *fp = zigzagStack.back();
    zigzagStack.pop_back();
    int eind=-1;
    if(CountSelectedEdges(fp,eind)==2)
    {
      fp->FEp(eind)->SetS();
      fp->FEp((eind+1)%3)->ClearS();
      fp->FEp((eind+2)%3)->ClearS();
      zigzagStack.push_back(fp->FFp(eind));
    }
    if(cnt++%5000 == 0) Dump(m,"smoothing%05i.ply",cnt);
  }
  Dump(m,"smoothed.ply",0);

  // Step 5)
  // Actually Cutting the mesh
  // We start from a border and we walk over it remainng always on the same side.
  // During the walk we add to all the wedges of the touched faces the index of a new vertex
  // At the end of the walk we actually add vertices and substitute the vertex references
  // in all the faces using the indexes that we have previously computed


  tri::UpdateFlags<MyMesh>::VertexClearV(m);
  face::Pos<MyFace> startPos;
  // Find a first face with a selected edge
  for(MyMesh::FaceIterator fi=m.face.begin();fi!=m.face.end();fi++)
  {
    for(int i=0;i<3;++i)
      if(fi->FEp(i)->IsS())
      {
        startPos =  face::Pos<MyFace>(&*fi,i,fi->V(i));
         assert(startPos.F()->FEp(startPos.E())->IsS());
      }
  }

  MyMesh::PerFaceAttributeHandle<NewVertInd> nvH =
      tri::Allocator<MyMesh>::AddPerFaceAttribute< NewVertInd > (m);

  // Walk along the border...
  int nvi =0;
  face::Pos<MyFace> curPos=startPos;
  do
  {
    assert(curPos.F()->FEp(curPos.E())->IsS());
    curPos.V()->SetV();
    do
    {
      nvH[curPos.F()].v[curPos.VInd()] = nvi;
      curPos.FlipE();
      curPos.FlipF();
    }
    while (!curPos.F()->FEp(curPos.E())->IsS());

    curPos.FlipF();
    curPos.FlipV();
    nvi++;
  }
  while(curPos!=startPos);
  // After the walk we actually add the vertexes and substitute the vertex ref in the faces.
  MyMesh::VertexIterator vi=tri::Allocator<MyMesh>::AddVertices(m,nvi);
  for(MyMesh::FaceIterator fi=m.face.begin();fi!=m.face.end();fi++)
  {
    for(int i=0;i<3;++i)
    {
      if(nvH[fi].v[i]>=0) // we touch only the vertex ref that we have previously filled
      {
        (vi+nvH[fi].v[i])->P() = fi->V(i)->P();
        fi->V(i)=&*(vi+nvH[fi].v[i]);
      }
    }
  }

  // DONE!!
  tri::io::ExporterPLY<MyMesh>::Save(m,"cut.ply",tri::io::Mask::IOM_VERTCOLOR|tri::io::Mask::IOM_VERTQUALITY);

  return 0;
}
